/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#ifdef cloudPass
    const float eyeAltitude = 800.0;
#else
    uniform float eyeAltitude;
#endif

uniform float wetness;

uniform vec2 viewSize;

uniform vec3 sunDir;
uniform vec3 moonDir;

uniform vec4 daytime;

uniform sampler2D gaux2;

flat out mat4x3 lightColor;

#define airmassStepBias 0.33
#include "/lib/atmos/air/const.glsl"
#include "/lib/atmos/air/density.glsl"
#include "/lib/atmos/project.glsl"

void getColorPalette() {
    vec3 airEyePos = vec3(0.0, planetRad + eyeAltitude, 0.0);

    lightColor[0]  = getAirTransmittance(airEyePos, sunDir, 6) * sunIllum * sunlightIllum;

    #if !(defined cloudPass || defined skyboxPass)
        lightColor[0] *= (1.0 - wetness * 0.95);
    #endif
    
    #ifdef fogPass
        lightColor[1]  = texture(gaux2, projectSky(vec3(0.0, 1.0, 0.0))).rgb * skylightIllum;
    #elif (defined cloudPass)
        lightColor[1]  = texture(gaux2, projectSky(vec3(0.0, 1.0, 0.0))).rgb * pi * 0.25;
    #else
        lightColor[1]  = texture(gaux2, projectSky(vec3(0.0, 1.0, 0.0))).rgb * pi * skylightIllum;
    #endif

    lightColor[1] *= vec3(skylightRedMult, skylightGreenMult, skylightBlueMult);

    lightColor[2]  = getAirTransmittance(airEyePos, moonDir, 6) * moonIllum * moonlightIllum;
        
    #ifndef skipBlocklight
        lightColor[3]  = blackbody(float(blocklightBaseTemp)) * blocklightIllum * blocklightBaseMult;
    #endif
}