#include "/lib/offset/gauss.glsl"

#define atrousDepthTreshold 0.001
#define atrousDepthOffset 0.2
#define atrousDepthExp 8.0

#define atrousNormalExponent 16

/*
luma exp: smol = sharper
luma offset: big = moar min blur
*/

//#define atrousLumaOffset 8.0

mat2 computeVariance(usampler2D tex, ivec2 pos) {
    vec2 sumMsqr    = vec2(0.0);
    vec2 sumMean    = vec2(0.0);

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos     = kernelO_5x5[i];

        mat2x3 col = fetchLighting(tex, clampTexelPos(pos + deltaPos));
        vec2 lum   = vec2(avgOf(col[0]), avgOf(col[1]));

        sumMsqr    += sqr(lum);
        sumMean    += lum;
    }
    sumMsqr  /= 9.0;
    sumMean  /= 9.0;

    vec2 variance  = abs(sumMsqr - sqr(sumMean)) * rcp(max(sumMean, 1e-20));

    return mat2(variance.x, sumMean.x, variance.y, sumMean.y);
}

mat4x3 computeVarianceColor(usampler2D tex, ivec2 pos) {
    mat2x3 sumMsqr    = mat2x3(0.0);
    mat2x3 sumMean    = mat2x3(0.0);

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos     = kernelO_5x5[i];

        mat2x3 col = fetchLighting(tex, clampTexelPos(pos + deltaPos));

        sumMsqr[0]    += sqr(col[0]);
        sumMsqr[1]    += sqr(col[1]);
        sumMean[0]    += col[0];
        sumMean[1]    += col[1];
    }
    sumMsqr  /= 9.0;
    sumMean  /= 9.0;

    vec3 varianceA  = abs(sumMsqr[0] - sqr(sumMean[0])) * rcp(max(sumMean[0], 1e-20));
    vec3 varianceB  = abs(sumMsqr[1] - sqr(sumMean[1])) * rcp(max(sumMean[1], 1e-20));

    return mat4x3(varianceA, sumMean[0], varianceB, sumMean[1]);
}

#if 1
mat2x3 atrousSVGF(usampler2D tex, sampler2D gData, vec2 coord, const int size) {
    ivec2 pos           = ivec2(coord * viewSize);

    vec4 centerData     = texelFetch(gData, pos, 0);
        centerData.xyz  = centerData.xyz * 2.0 - 1.0;
        centerData.w    = (centerData.w) * far;

    mat2x3 centerColor  = fetchLighting(tex, pos);

    //return centerColor;

    mat4x3 variance     = computeVarianceColor(tex, pos);

    vec2 sigmaOffset    = vec2(0.16, 0.16);

    vec3 sigmaL_A       = rcp(sigmaOffset.x + pi * sqrt(variance[0]));
    vec3 sigmaL_B       = rcp(sigmaOffset.y + pi * sqrt(variance[2]));

    vec2 maxLumDelta    = vec2(pi, euler);

    mat2x3 total        = centerColor;
    vec3 totalWeightA   = vec3(1.0);
    vec3 totalWeightB   = vec3(1.0);

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];
        if (deltaPos.x == 0 && deltaPos.y == 0) continue;

        ivec2 samplePos     = pos + deltaPos * size;

        bool valid          = all(greaterThanEqual(samplePos, ivec2(0))) && all(lessThan(samplePos, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = texelFetch(gData, clampTexelPos(samplePos), 0);
            currentData.xyz = currentData.xyz * 2.0 - 1.0;
            currentData.w   = (currentData.w) * far;

        float depthDelta    = abs(currentData.w - centerData.w) * atrousDepthExp;

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), atrousNormalExponent * 1.0);
            spatialWeight  *= exp(-depthDelta);

        //if (weight < 1e-20) continue;

        mat2x3 currentColor = fetchLighting(tex, samplePos);

        vec3 lumaDelta_A    = abs(centerColor[0] - currentColor[0]) / clamp(mix(variance[1], centerColor[0], 0.5), 1e-2, 2e2);
        vec3 lumaDelta_B    = abs(centerColor[1] - currentColor[1]) / clamp(variance[3], 1e-2, 2e2);

        vec3 weight_A       = exp(-clamp(sqr(lumaDelta_A), 0.0, maxLumDelta.x)) * spatialWeight;
        vec3 weight_B       = exp(-clamp(lumaDelta_B, 0.0, maxLumDelta.y)) * spatialWeight;

        //accumulate stuff
        total[0]    += currentColor[0] * weight_A;
        total[1]    += currentColor[1] * weight_B;

        totalWeightA += weight_A;
        totalWeightB += weight_B;
    }

    //compensate for total sampling weight
    total[0]       *= rcp(max(totalWeightA, 1e-25));
    total[1]       *= rcp(max(totalWeightB, 1e-25));

    return total;
}
#else

#define atrousLumaExp 3.0

mat2x3 atrousSVGF(usampler2D tex, sampler2D gData, vec2 coord, const int size) {
    ivec2 pos           = ivec2(coord * viewSize);

    vec4 centerData     = texelFetch(gData, pos, 0);
        centerData.xyz  = centerData.xyz * 2.0 - 1.0;
        centerData.w    = (centerData.w) * far;

    mat2x3 centerColor  = fetchLighting(tex, pos);
    vec2 centerLuma     = vec2(getLuma(centerColor[0]), getLuma(centerColor[1]));

    mat2 variance       = computeVariance(tex, pos);

    vec2 sigmaOffset    = vec2(0.71, 0.16);

    const vec2 lumaDeltaMult = 1.0 / vec2(euler, sqrt2);

    vec2 sigmaL         = rcp(sigmaOffset + vec2(euler, sqrt2) * vec2(variance[0].x, variance[1].x));
    vec2 maxLumDelta    = vec2(pi, euler);

    mat2x3 total        = centerColor;
    vec2 totalWeight    = vec2(1.0);

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];
        if (deltaPos.x == 0 && deltaPos.y == 0) continue;

        ivec2 samplePos     = pos + deltaPos * size;

        bool valid          = all(greaterThanEqual(samplePos, ivec2(0))) && all(lessThan(samplePos, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = texelFetch(gData, clampTexelPos(samplePos), 0);
            currentData.xyz = currentData.xyz * 2.0 - 1.0;
            currentData.w   = (currentData.w) * far;

        float depthDelta    = abs(currentData.w - centerData.w) * atrousDepthExp;

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), atrousNormalExponent * 1.0);
            spatialWeight  *= exp(-depthDelta);

        //if (weight < 1e-20) continue;

        mat2x3 currentColor = fetchLighting(tex, samplePos);
        vec2 currentLuma    = vec2(getLuma(currentColor[0]), getLuma(currentColor[1]));

        vec2 lumaDelta      = abs(centerLuma - currentLuma) / clamp(vec2(variance[0].y, variance[1].y), 1e-2, 2e2);

        vec2 weight         = exp(-clamp(lumaDelta, vec2(0.0), maxLumDelta) / lumaDeltaMult) * spatialWeight;

        //accumulate stuff
        total[0]    += currentColor[0] * weight.x;
        total[1]    += currentColor[1] * weight.y;

        totalWeight += weight;
    }

    //compensate for total sampling weight
    total[0]       *= rcp(max(totalWeight.x, 1e-25));
    total[1]       *= rcp(max(totalWeight.y, 1e-25));

    return total;
}
#endif