/*
    Brightness in 10Lux
*/

#define LIGHT_TORCH vec3(1.0, 0.55, 0.18)
#define LIGHT_REDTORCH vec3(1.0, 0.3, 0.1)
#define LIGHT_SOUL vec3(0.26, 0.42, 1.0)
#define LIGHT_FIRE vec3(1.0, 0.45, 0.12)
#define LIGHT_ENDROD vec3(0.76, 0.35, 1.0)

vec3 getEmission(int ID, vec3 albedo) {
    float lum   = getLuma(albedo);
    //albedo  = saturate(mix(cube(lum) * normalize(albedo), albedo, lum));

    float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
        albedoLum   = saturate(albedoLum * sqrt2);

    float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

    albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;

    if (ID == 1) return albedo * 200;

    if (ID == 2) return albedo * 160;

    if (ID == 3) return albedo * 80;

    if (ID == 10) return LIGHT_TORCH * 120;

    if (ID == 11) return LIGHT_REDTORCH * 35;

    if (ID == 12) return LIGHT_SOUL * 60;

    if (ID == 13) return LIGHT_FIRE * 80;

    if (ID == 14) return LIGHT_ENDROD * 100;

    return vec3(0.0);
}

vec3 getEmissionScreenSpace(int ID, vec3 albedo, float textureEmission) {

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);
            emitterLum  = max(emitterLum, textureEmission);

        //albedo  = saturate(mix(cube(lum) * normalize(albedo), albedo, lum));
        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    if (ID == 1) return albedo * 200;

    if (ID == 2) return albedo * 160;

    if (ID == 3) return albedo * 80;

    if (ID == 10) return LIGHT_TORCH * 120 * emitterLum;

    if (ID == 11) return LIGHT_REDTORCH * 35 * emitterLum;

    if (ID == 12) return LIGHT_SOUL * 60 * emitterLum;

    if (ID == 13) return LIGHT_FIRE * 80 * emitterLum;

    if (ID == 14) return LIGHT_ENDROD * 100 * emitterLum;

    #if labEmissionMode != 2
        return albedo * textureEmission * 80;
    #else
        return albedo * 80;
    #endif
}