bool IntersectAABox(vec3 rayPos, vec3 invRayDir, vec3 boxRadii, out float dist, inout vec3 hitNormal) {
	// This one will not intersect backfaces, or in other words this will not intersect the box from inside it.
	vec3 scaledRayPos = rayPos * invRayDir;
	vec3 minDists = -boxRadii * abs(invRayDir) - scaledRayPos; // distance to each frontfacing plane
	vec3 maxDists =  boxRadii * abs(invRayDir) - scaledRayPos; // distance to each backfacing plane

	float maxOfMin = maxOf(minDists); // distance to the frontfaces
	float minOfMax = minOf(maxDists); // distance to the backfaces

	dist = maxOfMin; // Distance to intersection (if the box was intersected)

	// Determine normal
	if (dist == minDists.x) {
		hitNormal = vec3(invRayDir.x < 0.0 ? 1.0 : -1.0, 0.0, 0.0);
	} else if (dist == minDists.y) {
		hitNormal = vec3(0.0, invRayDir.y < 0.0 ? 1.0 : -1.0, 0.0);
	} else {
		hitNormal = vec3(0.0, 0.0, invRayDir.z < 0.0 ? 1.0 : -1.0);
	}

	// Check for intersection & return whether or not we intersected
	return dist > 0.0 && maxOfMin <= minOfMax;
}
bool IntersectAABox(vec3 rayPos, vec3 invRayDir, vec3 minBounds, vec3 maxBounds, out float dist, inout vec3 hitNormal) {
	// This will not intersect backfaces, or in other words this will not intersect the box from inside it.
	// You should use the version that rakes in radii instead of min & max bounds if you can, as it will be faster.
	vec3 minBoundsDist = (minBounds - rayPos) * invRayDir;
	vec3 maxBoundsDist = (maxBounds - rayPos) * invRayDir;

	vec3 minDists = min(minBoundsDist, maxBoundsDist);
	vec3 maxDists = max(minBoundsDist, maxBoundsDist);

	float maxOfMin = maxOf(minDists); // distance to frontfaces
	float minOfMax = minOf(maxDists); // distance to backfaces

	dist = maxOfMin; // Distance to intersection (if the box was intersected)

	// Determine normal
	if (dist == minDists.x) {
		hitNormal = vec3(invRayDir.x < 0.0 ? 1.0 : -1.0, 0.0, 0.0);
	} else if (dist == minDists.y) {
		hitNormal = vec3(0.0, invRayDir.y < 0.0 ? 1.0 : -1.0, 0.0);
	} else {
		hitNormal = vec3(0.0, 0.0, invRayDir.z < 0.0 ? 1.0 : -1.0);
	}

	// Check for intersection & return whether or not we intersected
	return dist > 0.0 && maxOfMin <= minOfMax;
}

bool intersectVoxel(vec3 relOrigin, vec3 dir, int ID, inout float hitDist, inout vec3 hitNormal) {
    vec3 invDir     = 1.0 / dir;

    vec3 minBounds  = vec3(0);
    vec3 maxBounds  = vec3(1);

    if (ID == 1) {              //bottom slabs
        minBounds  = vec3(0.0, 0.0, 0.0);
        maxBounds  = vec3(1.0, 0.5, 1.0);
    } else if (ID == 2) {       //top slabs
        minBounds  = vec3(0.0, 0.5, 0.0);
        maxBounds  = vec3(1.0, 1.0, 1.0);
    } else if (ID == 3) {       //pressure plates
        minBounds  = vec3(0.0625, 0, 0.0625);
        maxBounds  = vec3(0.9375, 0.0625, 0.9375);
    } else if (ID == 4) {       //carpets
        minBounds  = vec3(0.0, 0.0, 0.0);
        maxBounds  = vec3(1.0, 0.0625, 1.0);
    } else if (ID == 10) {      //torches
        minBounds  = vec3(0.4, 0.125, 0.4);
        maxBounds  = vec3(0.6, 0.625, 0.6);
    } else if (ID == 20) {      //End Rod Up-Down
        minBounds  = vec3(0.45, 0.05, 0.45);
        maxBounds  = vec3(0.55, 0.95, 0.55);
    } else if (ID == 21) {      //End Rod East-West
        minBounds  = vec3(0.05, 0.45, 0.45);
        maxBounds  = vec3(0.95, 0.55, 0.55);
    } else if (ID == 22) {      //End Rod North-South
        minBounds  = vec3(0.45, 0.45, 0.05);
        maxBounds  = vec3(0.55, 0.55, 0.95);
    } else if (ID == 23) {      // Trapdoor Bottom
        minBounds  = vec3(0.0, 0.0, 0.0);
        maxBounds  = vec3(1.0, 0.15, 1.0);
    } else if (ID == 24) {      // Door East Closed
        minBounds  = vec3(0.0, 0.0, 0.0);
        maxBounds  = vec3(0.15, 1.0, 1.0);
    } else if (ID == 25) {      // Door West Closed
        minBounds  = vec3(0.85, 0.0, 0.0);
        maxBounds  = vec3(1.0, 1.0, 1.0);
    } else if (ID == 26) {      // Door South Closed
        minBounds  = vec3(0.0, 0.0, 0.0);
        maxBounds  = vec3(1.0, 1.0, 0.15);
    } else if (ID == 27) {      // Door North Closed
        minBounds  = vec3(0.0, 0.0, 0.85);
        maxBounds  = vec3(1.0, 1.0, 1.0);
    } else if (ID == 28) {      // Trapdoor Top
        minBounds  = vec3(0.0, 0.85, 0.0);
        maxBounds  = vec3(1.0, 1.0, 1.0);
    } else {
        return true;
    }

    bool intersected = IntersectAABox(relOrigin, invDir, minBounds, maxBounds, hitDist, hitNormal);

    return intersected;
}