/*
Voxel storage structure based off VXPT by Zombye
*/

uniform float eyeAltitude;

vec3 sceneToVoxelSpace(vec3 pos) {
    const float radius  = voxelMapRes / 32;

	pos.xz += fract(cameraPosition.xz) + radius;    // Center around player
	pos.y  += eyeAltitude;
    pos.y  -= clamp(-64.0 + round(eyeAltitude), 0.0, 128.0);

    return pos;
}
vec3 voxelToSceneSpace(vec3 pos) {
    const float radius  = voxelMapRes / 32;

	pos.xz -= fract(cameraPosition.xz) + radius;
	pos.y  -= eyeAltitude;
    pos.y  += clamp(-64.0 + round(eyeAltitude), 0.0, 128.0);

	return pos;
}

bool withinVoxelVolume(uvec3 index) {
    const uvec2 maxBounds   = uvec2(voxelMapRes / 16, 128) - 1u;

    return all(lessThanEqual(index, maxBounds.xyx));
}
bool withinVoxelVolume(ivec3 index) {
    return withinVoxelVolume(uvec3(index));
}
bool withinVoxelVolume(vec3 index) {
    const vec2 maxBounds   = vec2(voxelMapRes / 16.0, 128.0) - 1.0;

    return all(greaterThanEqual(index, vec3(0.0))) && all(lessThanEqual(uvec3(index), maxBounds.xyx));
}

bool outsideVoxelVolume(uvec3 index) {
    const uvec2 maxBounds   = uvec2(voxelMapRes / 16, 128) - 1u;

    return any(greaterThan(index, maxBounds.xyx));
}
bool outsideVoxelVolume(ivec3 index) {
    return outsideVoxelVolume(uvec3(index));
}
bool outsideVoxelVolume(vec3 index) {
    const vec2 maxBounds   = vec2(voxelMapRes / 16, 128) - 1.0;

    return any(lessThan(index, vec3(0.0))) || any(greaterThan(index, maxBounds.xyx));
}

uvec2 getVoxelIndex(uvec3 index) {
    const uint radius   = uint(voxelMapRes / 16);

    uvec2 tileIndex     = uvec2(index.y & 15u, index.y >> 4u);
    uvec2 pos           = index.xz + radius * tileIndex;

    return pos;
}
ivec2 getVoxelIndex(ivec3 index) {
    const int radius    = voxelMapRes / 16;

    ivec2 tileIndex     = ivec2(index.y & 15, index.y >> 4);
    ivec2 pos           = index.xz + radius * tileIndex;

    return pos;
}
vec2 getVoxelCoord(vec3 index) {
    const float radius  = float(voxelMapRes / 16);

    const vec2 tiles    = vec2(16.0, 16.0);

    vec2 tileIndex;
        tileIndex.x     = modf(index.y / tiles.x, tileIndex.y) * tiles.x;

    vec2 pos            = index.xz + radius * tileIndex;

        pos    /= voxelMapRes;
        pos    += 0.5 / voxelMapRes;

    return pos;
}

vec3 uvToVoxelIndex(vec2 uv) {
    const float radius  = float(voxelMapRes / 16);
    const ivec2 tiles   = ivec2(16, 16);

    vec2 pos            = uv - 0.5 / voxelMapRes;
        pos            *= tiles;

    vec2 tileIndex      = floor(pos) * vec2(1.0, tiles.y);
    vec2 tileVoxel      = fract(pos) * radius;

    return vec3(tileVoxel.x, tileIndex.x + tileIndex.y, tileVoxel.y);
}

vec2 getVoxelCoord(vec3 index, int level) {
    int size        = int(pow(2, level));

        index      /= size;
        index       = floor(index);

    int mapRes      = voxelMapRes / size;
    float radius    = float(mapRes / 16);
    vec2 tiles      = vec2(16.0, 16.0);

    vec2 tileIndex;
        tileIndex.x     = modf(index.y / tiles.x, tileIndex.y) * tiles.x;

    vec2 pos            = index.xz + radius * tileIndex;

        pos    /= voxelMapRes;
        pos    += 0.5 / voxelMapRes;

    return pos;
}
vec3 uvToVoxelIndex(vec2 uv, int level) {
    int size        = int(pow(2, level));

    int mapRes      = voxelMapRes / size;

    float radius    = float(mapRes / 16);

    const ivec2 tiles   = ivec2(16, 16);

    vec2 pos            = uv - 0.5 / mapRes;
        pos            *= tiles;

    vec2 tileIndex      = floor(pos) * vec2(1.0, tiles.y);
    vec2 tileVoxel      = fract(pos) * radius;

    return vec3(tileVoxel.x, tileIndex.x + tileIndex.y, tileVoxel.y);
}
bool withinVoxelVolume(ivec3 index, int level) {
    int size        = int(pow(2, level));
    int mapRes      = voxelMapRes / size;

    ivec2 maxBounds   = ivec2(mapRes / 16, 128) - 1;

    return all(lessThanEqual(index, maxBounds.xyx));
}