/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 15 */
layout(location = 0) out vec3 smoothColor;

#include "/lib/head.glsl"

uniform sampler2D colortex3;
uniform sampler2D colortex15;

uniform float far, near;

uniform vec2 viewSize;

in vec2 coord;

void main() {
    smoothColor     = vec3(0.0);

    #ifdef ACCUM_ADAPTIVE
    vec2 uvScaled   = coord * indirectResScale * 4.0;

    if (saturate(uvScaled) == uvScaled) {
        ivec2 pixelUV   = ivec2(floor(coord * viewSize));

        vec4 normalDepthCenter  = texelFetch(colortex3, pixelUV, 0);
            normalDepthCenter.rgb = normalDepthCenter.rgb * 2.0 - 1.0;
            normalDepthCenter.a = sqr(normalDepthCenter.a) * far;

        vec3 avgColor   = texelFetch(colortex15, pixelUV, 0).rgb;
        const int r     = 3;
        float weight    = 1.0;

        for (int x = -r; x <= r; x++) {
            for (int y = -r; y <= r; y++) {
                if (x == 0 && y == 0) continue;

                ivec2 UV   = pixelUV + ivec2(x, y) * SVGF_SIZE;

                float w = float(all(greaterThanEqual(UV, ivec2(0)))
                        && all(lessThan(UV, ivec2(viewSize * indirectResScale / 4.0))));

                vec4 normalDepth = texelFetch(colortex3, UV, 0);
                    normalDepth.rgb = normalDepth.rgb * 2.0 - 1.0;
                    normalDepth.a   = sqr(normalDepth.a) * far;

                float distDepth = abs(normalDepthCenter.a - normalDepth.a) / abs(normalDepthCenter.a);

                    w      *= pow(max(0.0, dot(normalDepthCenter.rgb, normalDepth.rgb)), 4.0);
                    w      *= exp(-distDepth * 4.0);

                avgColor   += texelFetch(colortex15, UV, 0).rgb * w;
                weight     += w;
            }
        }

        smoothColor     = clamp16F(avgColor / weight);
    }
    #endif
}