#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"
#include "/lib/shadowconst.glsl"

layout(location = 0) out vec4 data0;
layout(location = 1) out vec4 data1;

#define gSHADOW

uniform sampler2D gcolor;

flat in int isVoxel;

flat in int matID;

flat in uvec4 voxelWrite;

in vec2 coord;

in vec3 tint;

in float warp;

in vec3 scenePos;
in vec3 worldPos;


/* ------ caustics ------ */

#define FUTIL_MAT16
#include "/lib/fUtil.glsl"

uniform sampler2D noisetex;

uniform float frameTimeCounter;

uniform vec3 lightDir;

#include "/lib/util/bicubic.glsl"

const float viewDist = 0.0;

#include "/lib/atmos/waterConst.glsl"

#include "/lib/atmos/waterWaves.glsl"

vec3 waterNormal(vec3 position, float dstep, int matID) {
        dstep       = max(dstep, 0.015);

    vec2 delta;
        delta.x     = waterWaves(position + vec3( dstep, 0.0, -dstep), matID);
        delta.y     = waterWaves(position + vec3(-dstep, 0.0,  dstep), matID);
        delta      -= waterWaves(position + vec3(-dstep, 0.0, -dstep), matID);

    return normalize(vec3(-delta.x, 2.0 * dstep, -delta.y));
}

float projectedCaustic(vec3 pos, vec3 normal, vec3 lightDir) {
    vec3 dPdx   = dFdx(pos);
    vec3 dPdy   = dFdy(pos);

    float num   = dotSelf(dPdx) * dotSelf(dPdy);

    vec3 refractLight = refract(-lightDir, normal, rcp(1.33));
        dPdx   += 2.0 * dFdx(refractLight);
        dPdy   += 2.0 * dFdy(refractLight);

    float denom = dotSelf(dPdx) * dotSelf(dPdy);

    return sqrt(num * rcp(denom));
}

void main() {
    if (matID == 31) discard;

    if (isVoxel == 1) {

        data0   = vec4(voxelWrite) / 65535.0;
        data1   = vec4(1.0, 0.0, 0.0, 0.0);

    } else {
        if (matID == 102 || matID == 103) { 
            vec3 waves      = waterNormal(worldPos, rcp(shadowMapResolution) * warp * 2.0, matID);
            vec3 refractLight = refract(lightDir, waves, rcp(1.33));
            float caustic   = projectedCaustic(scenePos, waves, lightDir);
            data0           = vec4(caustic * 0.25, 1.0, 1.0, 1.0);
        } else {
            data0           = texture(gcolor, coord, -shadowmapTextureMipBias);
            if (data0.a<0.1) discard;
                data0.rgb  *= tint.rgb;

            data0.rgb       = linearToAP1Albedo(toLinear(data0.rgb));
        }

        data1   = vec4(float(matID == 102 || matID == 103), 0.0, 0.0, 0.0);
    }
}