#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

layout(location = 0) out vec4 data0;
layout(location = 1) out vec4 data1;

in vec2 coord;

uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform vec3 cameraPosition;

#include "/lib/voxel/store.glsl"

const ivec3 sampleOffsets[8] = ivec3[8](
    ivec3(0, 0, 0), ivec3(1, 0, 0), ivec3(0, 0, 1), ivec3(1, 0, 1),
    ivec3(0, 1, 0), ivec3(1, 1, 0), ivec3(0, 1, 1), ivec3(1, 1, 1)
);

bool isVoxelOccupied(ivec3 index, int level) {
    if (!withinVoxelVolume(index)) return false;    // For obvious reasons

    return texture(shadowcolor1, getVoxelCoord(index, level))[level] > 0.5;
}

void main() {
    data0   = stex(shadowcolor0);   // Pass through shadowmap data, RENDERTARGETS doesn't seem to work here
    data1   = stex(shadowcolor1);

    const int size  = int(pow(2, CURR_LEVEL));
    const int size1  = int(pow(2, CURR_LEVEL-1));

    vec2 lodUV  = coord * vec2(size, (size));

    if (lodUV.y < 0.5 && lodUV.x <= 1.0) {

        vec3 index      = floor(uvToVoxelIndex(lodUV, CURR_LEVEL));

        bool anyVoxel   = false;

        for (int i = 0; i < 8; i++) {
            ivec3 offset    = sampleOffsets[i];

            if (isVoxelOccupied((ivec3(index) * 2 + offset) * size1, CURR_LEVEL - 1)) {   // Read from one level above the current one
                anyVoxel    = true;
                break;
            }
        }

        #if CURR_LEVEL == 1
            data1.y = float(anyVoxel);
        #elif CURR_LEVEL == 2
            data1.z = float(anyVoxel);
        #elif CURR_LEVEL == 3
            data1.w = float(anyVoxel);
        #endif
    }
}