#version 430 compatibility


/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



/* RENDERTARGETS: 3,7,15 */
layout(location = 0) out vec4 GData;
layout(location = 1) out vec3 currentFrame;
layout(location = 2) out vec3 albedoData;

#include "/lib/head.glsl"

uniform sampler2D colortex0;
uniform usampler2D colortex1;
uniform sampler2D colortex3;
uniform usampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;

uniform int frameCounter;
uniform int frame255;
uniform int frame255Last;

uniform float far, near;

uniform float sunAngle;

uniform vec2 viewSize;
uniform vec2 taaOffset;
uniform vec2 skyCaptureResolution;

uniform vec3 cameraPosition;
uniform vec3 sunDir, moonDir, lightDir;
uniform vec3 sunDirView, moonDirView, lightDirView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

in vec2 coord;

flat in mat4x3 lightColor;

#define gFragment

int atlasMipmapLevel = textureQueryLevels(colortex3)-1;
ivec2 atlasSize      = ivec2(textureSize(colortex3, 0));

#include "/lib/atmos/project.glsl"

#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"
//#include "/lib/voxel/octree.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"

#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/mat/emission.glsl"

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo), voxelID);
}

#define useMipBounceAlbedo

Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(colortex3, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(colortex3, ivec2(uv * atlasSize), 0);
    #endif

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y);
}

vec2 bluenoiseLookup(ivec2 pos) {
    ivec2 xy    = pos & 255;
    uint z      = frameCounter & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 bluenoiseLookup(ivec2 pos, int zIndex) {
    ivec2 xy    = pos & 255;
    uint z      = zIndex & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 bluenoiseLookupGrad(ivec2 pos) {
    ivec2 xy    = (pos + frameCounter) & 255;
    uint z      = frameCounter & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 unprojectSphere(vec3 dir) {
    vec2 lonlat     = vec2(atan(-dir.x, dir.z), acos(dir.y));
    return lonlat * vec2(rcp(tau), rcp(pi)) + vec2(0.5, 0.0);
}

float depthLinear(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

#include "/lib/light/warp.glsl"

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightDir;
    float a     = length(pos);
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}

vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
    float sDotN     = dot(lightDir, sceneNormal);
    if (sDotN <= 0.0) return vec3(0.0);

    vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);

    float shadow1   = texture(shadowtex1, shadowPos);

    return texture(shadowcolor0, shadowPos.xy).rgb * 4.0 * (shadow1 * saturate(sDotN)) / pi;
}


#define HASHSCALE1 .1031
float hash12(vec2 x) {
    vec3 x3  = fract(vec3(x.xyx) * HASHSCALE1);
        x3  += dot(x3, x3.yzx + 19.19);
    return fract((x3.x + x3.y) * x3.z);
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}

float ditherHashNoise() {
    float noise     = hash12(gl_FragCoord.xy);
        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)) *0);

    return noise;
}

#define rtHashNoise

float lambertianDiffuse(vec3 normal, vec3 direction) {
    /*
        This is used for some optimized or approximated stuff where Oren-Nayar gives very little visual advantages
        but a big performance loss instead.
    */
    float lambert   = saturate(dot(normal, direction));

    return lambert / pi;
}

uint initRand(uint val0, uint val1, uint backoff) {
    backoff = 16;
	uint v0 = val0, v1 = val1, s0 = 0;

	for (uint n = 0; n < backoff; n++)
	{
		s0 += 0x9e3779b9;
		v0 += ((v1 << 4) + 0xa341316c) ^ (v1 + s0) ^ ((v1 >> 5) + 0xc8013ea4);
		v1 += ((v0 << 4) + 0xad90777d) ^ (v0 + s0) ^ ((v0 >> 5) + 0x7e95761e);
	}
	return v0;
}

// Takes our seed, updates it, and returns a pseudorandom float in [0..1]
float nextRand(inout uint s) {
	s = (1664525u * s + 1013904223u);
	return float(s & 0x00FFFFFF) / float(0x01000000);
}

vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}

vec3 genCosineVector(vec3 vector, vec3 xyz) {
	vec3 dir = genUnitVector(xyz.xy);
	float VoD = dot(vector, dir);

	return sqrt(xyz.z) * (dir - vector * VoD) * inversesqrt(1.0 - VoD * VoD) + sqrt(1.0 - xyz.z) * vector;
}

vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 10;
    const float stepSize    = halfPi / float(maxSteps);

    vec3 stepVector         = direction * stepSize;

    vec3 endPosition        = position + stepVector * maxSteps;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec2 maxPosXY           = max(abs(endScreenPosition.xy * 2.0 - 1.0), vec2(1.0));
    float stepMult          = minOf(vec2(1.0) / maxPosXY);
        stepVector         *= stepMult;

    // closest texel iteration
    vec3 samplePos          = position;
    
        samplePos          += stepVector / 6.0;
    vec3 screenPos          = viewToScreenSpace(samplePos);

    const float thresh      = 0.25;

    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }
    
        samplePos          += stepVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        vec3 screenPos      = viewToScreenSpace(samplePos);
            samplePos      += stepVector;
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }

    return vec3(1.1);
}

mat2x3 unpackDoubleRGBE8(uvec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

bool screenRaytrace(vec3 pos, vec3 direction, inout vec3 hitPosition, inout vec3 hitIndex, inout vec3 hitNormal, inout Material hitMat, inout vec3 light) {
        pos     = voxelToSceneSpace(pos);
        pos     = viewToSceneSpace(pos, gbufferModelView);

        direction   = mat3(gbufferModelView) * direction;
    
    vec3 hitPos     = screenspaceRT(pos, direction, ditherHashNoise());

    if (hitPos.z >= 1.0) return false;

    ivec2 pixelPos  = ivec2(hitPos.xy * viewSize);

    uvec4 tex1      = texelFetch(colortex1, pixelPos, 0);

    mat2x3 hitColor = unpackDoubleRGBE8(texelFetch(colortex4, pixelPos, 0));

    vec3 albedo     = texelFetch(colortex0, pixelPos, 0).rgb;

        hitNormal   = decodeNormal(invF16UI(tex1.xy));

    hitMat      = Material(hitNormal, albedo, hitColor[1], 0);
    light       = hitColor[0];

    hitPos      = screenToViewSpace(hitPos, true);
    hitPos      = viewToSceneSpace(hitPos);

    hitPosition = sceneToVoxelSpace(hitPos);
    hitIndex    = floor(hitPosition - hitNormal / 128.0);

    return true;
}

float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
    return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 8.0) / 8.0));
}

mat2x3 PTLight(vec3 scenePos, vec3 sceneNormals, Material surfaceMat, uint randSeed, vec2 uv) {
    mat2x3 rayColor = mat2x3(0.0);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;

    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi * sqrt2;

    vec2 quasirandomCurr2 = 0.5 + fract(vec2(a1, a2) * frame255 + 0.5);
    
    vec3 startPos       = sceneToVoxelSpace(scenePos);

    mat2x3 lighting     = unpackDoubleRGBE8(texelFetch(colortex4, ivec2(uv * viewSize), 0));
    bool inSunlight     = avgOf(lighting[0] * pi) > rpi;

    for (int rayIndex = 0; rayIndex < PT_SPP; ++rayIndex) {
        int frame255new     = frame255 + int(rayIndex * 256.0);

        vec2 randSinCos     = vec2(cos(quasirandomCurr2.x), sin(quasirandomCurr2.y));

        vec2 noiseCurr  = hash22(gl_FragCoord.xy + frame255new);

        vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frame255new + 0.5);

        vec3 contribution = vec3(1.0);

        Material hitMat = surfaceMat;

        vec3 hitNormal = hitMat.normal;

        vec3 voxelPos   = sceneToVoxelSpace(scenePos);
        vec3 voxelIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 startIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 cameraPos  = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

        vec3 direction  = normalize(voxelPos - cameraPos);

        bool hitSky     = false;

        //++quasirandomCurr2;

        for (int i = 0; i < bounceNumber; ++i) {
            if (i >= 2 && inSunlight) break;

            ++quasirandomCurr;

            noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

            bool initialVoxel   = false;
            vec3 oldDirection   = direction;

            vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

            #ifdef rtAlternateVector
                vec3 randVector = vec3(nextRand(randSeed), nextRand(randSeed), nextRand(randSeed));
                direction       = genCosineVector(hitNormal, randVector);
            #else
                direction       = GenerateCosineVectorSafe(hitNormal, vectorXY);
            #endif

            vec3 brdf = hitMat.albedo;
                brdf *= diffuseHammon(hitNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, hitNormal)/pi);

            contribution *= saturate(brdf);

            if(dot(hitNormal, direction) < 0.0 && !initialVoxel) continue;

            #ifdef SSPT_Enabled
                vec3 directLight = vec3(0.0);
                bool hit    = screenRaytrace(voxelPos, direction, voxelPos, voxelIndex, hitNormal, hitMat, directLight);

                if (hit) {
                    rayColor[0] += contribution * hitMat.emissive;

                    if (i >= 1) rayColor[1] +=  contribution * directCol * directLight;

                    hitSky      = false;

                    continue;
                }
            #else
                bool hit;
            #endif

            if(i >= 1) {
                rayColor[1] += contribution * directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) * hitMat.albedo;
            }

            #if OCTREE_LEVELS > 1
            hit     = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData);
            #else
            hit     = marchVoxel(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData);
            #endif

            if (hit) {
                hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, i);

                rayColor[0] += contribution * hitMat.emissive * getEmissionContribution(voxelPos, startPos);

                hitSky      = false;
            }


            if (!hit) {
                hitSky      = true;

                break;
            }
        }

        if (hitSky) {
            vec2 sphereCoord    = unprojectSphere(direction);
                sphereCoord    *= rcp(skyCaptureResolution);
                sphereCoord.y  += exp2(-SKY_RENDER_LOD) * 2.0;

            rayColor[1]   += contribution * texture(colortex5, sphereCoord).rgb * (sstep(direction.y, -0.15, -0.05) * 0.8 + 0.2);
        }
    }
    rayColor[0]     = max0(rayColor[0] / PT_SPP);
    rayColor[1]     = max0(rayColor[1] / PT_SPP) * sqrt2;

    return rayColor;
}

ivec2 clampTexelPos(ivec2 pos) {
    return clamp(pos, ivec2(0.0), ivec2(viewSize));
}

uvec4 packDoubleRGBE8(vec3 colorA, vec3 colorB) {
    vec4 RGBE_A     = encodeRGBE8(colorA);
    vec4 RGBE_B     = encodeRGBE8(colorB);

    return uvec4(pack2x8UI(RGBE_A.xy),
                pack2x8UI(RGBE_A.zw),
                pack2x8UI(RGBE_B.xy),
                pack2x8UI(RGBE_B.zw));
}

bool screenRaytrace(vec3 pos, vec3 direction, inout vec3 hitPosition, inout vec3 hitIndex, inout vec3 hitNormal, inout Material hitMat) {
        pos     = voxelToSceneSpace(pos);
        pos     = viewToSceneSpace(pos, gbufferModelView);

        direction   = mat3(gbufferModelView) * direction;
    
    vec3 hitPos     = screenspaceRT(pos, direction, ditherHashNoise());

    if (hitPos.z >= 1.0) return false;

    ivec2 pixelPos  = ivec2(hitPos.xy * viewSize);

    uvec4 tex1      = texelFetch(colortex1, pixelPos, 0);

    vec3 albedo     = texelFetch(colortex0, pixelPos, 0).rgb;

        hitNormal   = decodeNormal(invF16UI(tex1.xy));

    hitMat      = Material(hitNormal, albedo, vec3(0.0), 0);

    hitPos      = screenToViewSpace(hitPos, true);
    hitPos      = viewToSceneSpace(hitPos);

    hitPosition = sceneToVoxelSpace(hitPos);
    hitIndex    = floor(hitPosition - hitNormal / 128.0);

    return true;
}

vec3 PTAlbedo(vec3 scenePos, vec3 sceneNormals, Material surfaceMat, uint randSeed) {
    mat2x3 rayColor = mat2x3(0.0);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;

    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi * sqrt2;

    vec2 quasirandomCurr2 = 0.5 + fract(vec2(a1, a2) + 0.5);
    
    vec3 startPos       = sceneToVoxelSpace(scenePos);

    for (int rayIndex = 0; rayIndex < ADAPT_SPP; ++rayIndex) {
        int frame255new     = 1 + int(rayIndex * 256.0);

        vec2 randSinCos     = vec2(cos(quasirandomCurr2.x), sin(quasirandomCurr2.y));

        vec2 noiseCurr  = hash22(gl_FragCoord.xy + frame255new);

        vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frame255new + 0.5);

        vec3 contribution = vec3(1.0);

        Material hitMat = surfaceMat;

        vec3 hitNormal = hitMat.normal;

        vec3 voxelPos   = sceneToVoxelSpace(scenePos);
        vec3 voxelIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 startIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 cameraPos  = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

        vec3 direction  = normalize(voxelPos - cameraPos);

        bool hitSky     = false;

        //++quasirandomCurr2;

        for (int i = 0; i < ADAPT_BOUNCE; ++i) {
            ++quasirandomCurr;

            noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

            bool initialVoxel   = false;
            vec3 oldDirection   = direction;

            vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

            #ifdef rtAlternateVector
                vec3 randVector = vec3(nextRand(randSeed), nextRand(randSeed), nextRand(randSeed));
                direction       = genCosineVector(hitNormal, randVector);
            #else
                direction       = GenerateCosineVectorSafe(hitNormal, vectorXY);
            #endif

            vec3 brdf = hitMat.albedo;
                brdf *= diffuseHammon(hitNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, hitNormal)/pi);
                //brdf   *= saturate(dot(-oldDirection, hitNormal)) / saturate(dot(direction, hitNormal));
            contribution *= saturate(brdf);

            if(dot(hitNormal, direction) < 0.0 && !initialVoxel) continue;

            #ifdef SSPT_Enabled
                vec3 directLight = vec3(0.0);
                bool hit    = screenRaytrace(voxelPos, direction, voxelPos, voxelIndex, hitNormal, hitMat, directLight);

                if (hit) {
                    rayColor[0] += contribution * hitMat.emissive;

                    if (i >= 1) rayColor[1] +=  contribution * directCol * directLight;

                    hitSky      = false;

                    continue;
                }
            #else
                bool hit;
            #endif

            if(i >= 1) {
                rayColor[1] += contribution * directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) * hitMat.albedo;
            }

            #if OCTREE_LEVELS > 1
            hit     = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData);
            #else
            hit     = marchVoxel(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData);
            #endif
            
            if (hit) {
                hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, i);

                rayColor[0] += contribution * hitMat.emissive * getEmissionContribution(voxelPos, startPos);

                hitSky      = false;
            }


            if (!hit) {
                hitSky      = true;

                break;
            }
        }

        if (hitSky) {
            vec2 sphereCoord    = unprojectSphere(direction);
                sphereCoord    *= rcp(skyCaptureResolution);
                sphereCoord.y  += exp2(-SKY_RENDER_LOD) * 2.0;

            rayColor[1]   += contribution * texture(colortex5, sphereCoord).rgb * (sstep(direction.y, -0.15, -0.05) * 0.8 + 0.2);
        }
    }
    rayColor[0]     = max0(rayColor[0] / ADAPT_SPP);
    rayColor[1]     = max0(rayColor[1] / ADAPT_SPP) * sqrt2;

    return rayColor[0] + rayColor[1];
}

void main() {
    vec2 lowresCoord    = coord / indirectResScale;
    ivec2 pixelPos      = ivec2(floor(coord * viewSize - taaOffset * 0.5) / indirectResScale);
    float sceneDepth    = texelFetch(depthtex0, pixelPos, 0).x;

    currentFrame    = vec3(0);
    GData           = vec4(0.0, 0.0, 1.0, 1.0);

    uint randSeed = initRand(uint(gl_FragCoord.x + gl_FragCoord.y * viewSize.x), uint(frameCounter), 16);

    if (landMask(sceneDepth) && saturate(lowresCoord) == lowresCoord) {
        vec2 coord      = saturate(lowresCoord);
        uvec4 tex1      = texelFetch(colortex1, pixelPos, 0);

        ivec2 blockID   = unpack2x8I(tex1.z);

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth), true);
        vec3 scenePos   = viewToSceneSpace(viewPos);

        vec3 sceneNormal = decodeNormal(invF16UI(tex1.xy));

        //GData           = saturate(vec4(sceneNormal * 0.5 + 0.5, sqrt(depthLinear(sceneDepth) / far)));

        Material surfaceMat = Material(sceneNormal, vec3(1.0), vec3(0.0), blockID.y);

        mat2x3 result   = PTLight(scenePos, sceneNormal, surfaceMat, randSeed, lowresCoord);

        currentFrame    = clamp16F(result[0] + result[1]);
    }

    #ifdef ACCUM_ADAPTIVE
    ivec2 pixelUV       = ivec2(floor(coord * viewSize - taaOffset * 0.5) / indirectResScale) * 4;
        sceneDepth      = texelFetch(depthtex0, pixelUV, 0).x;
        lowresCoord     = (coord / indirectResScale) * 4.0;

    if (landMask(sceneDepth) && saturate(lowresCoord) == lowresCoord) {
        vec2 coord      = saturate(lowresCoord);
        uvec4 tex1      = texelFetch(colortex1, pixelUV, 0);

        ivec2 blockID   = unpack2x8I(tex1.z);

        //sceneDepth      = texture(depthtex0, coord - taaOffset * 0.5).x;

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth), true);
        vec3 scenePos   = viewToSceneSpace(viewPos);

        vec3 sceneNormal = texture(colortex15, coord).rgb * 2.0 - 1.0;

        GData           = saturate(vec4(sceneNormal * 0.5 + 0.5, sqrt(depthLinear(sceneDepth))));

        Material surfaceMat = Material(sceneNormal, vec3(1.0), vec3(0.0), blockID.y);

        albedoData      = PTAlbedo(scenePos, sceneNormal, surfaceMat, randSeed);
    }
    #endif
}