#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 7,9,10,11,14 */
layout(location = 0) out vec4 result;
layout(location = 1) out vec4 accumulated;
layout(location = 2) out vec4 GDataHistory;
layout(location = 3) out vec3 smoothColorStore;
layout(location = 4) out vec3 varianceStore;
//layout(location = 4) out vec3 variance;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform usampler2D colortex1;
uniform sampler2D colortex3;
uniform sampler2D colortex7;
uniform sampler2D colortex9;
uniform sampler2D colortex10;
uniform sampler2D colortex11;
uniform sampler2D colortex14;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

const bool colortex9Clear   = false;
const bool colortex10Clear  = false;
const bool colortex11Clear  = false;
const bool colortex14Clear  = false;

uniform float far, near;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousProjection, gbufferPreviousModelView;

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/util/transforms.glsl"

/* ------ reprojection ----- */
vec3 reproject(vec3 sceneSpace, bool hand) {
    vec3 prevScreenPos = hand ? vec3(0.0) : cameraPosition - previousCameraPosition;
    prevScreenPos = sceneSpace + prevScreenPos;
    prevScreenPos = transMAD(gbufferPreviousModelView, prevScreenPos);
    prevScreenPos = transMAD(gbufferPreviousProjection, prevScreenPos) * (0.5 / -prevScreenPos.z) + 0.5;

    return prevScreenPos;
}

uvec4 packDoubleRGBE8(vec3 colorA, vec3 colorB) {
    vec4 RGBE_A     = encodeRGBE8(colorA);
    vec4 RGBE_B     = encodeRGBE8(colorB);

    return uvec4(pack2x8UI(RGBE_A.xy),
                pack2x8UI(RGBE_A.zw),
                pack2x8UI(RGBE_B.xy),
                pack2x8UI(RGBE_B.zw));
}
uvec4 packDoubleRGBE8(mat2x3 color) {
    vec4 RGBE_A     = encodeRGBE8(color[0]);
    vec4 RGBE_B     = encodeRGBE8(color[1]);

    return uvec4(pack2x8UI(RGBE_A.xy),
                pack2x8UI(RGBE_A.zw),
                pack2x8UI(RGBE_B.xy),
                pack2x8UI(RGBE_B.zw));
}
mat2x3 unpackDoubleRGBE8(uvec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

mat2x3 fetchLighting(usampler2D tex, ivec2 coord) {
    return unpackDoubleRGBE8(texelFetch(tex, coord, 0));
}
mat2x3 mixMat2x3(mat2x3 a, mat2x3 b, float alpha) {
    mat2x3 x    = mat2x3(0.0);
        x[0]    = mix(a[0], b[0], alpha);
        x[1]    = mix(a[1], b[1], alpha);

    return x;
}

mat2x3 sampleLighting(usampler2D tex, vec2 uv) {
    vec2 pos        = uv * viewSize - 0.5;
    ivec2 pixelPos  = ivec2(pos);

    vec2 weights    = fract(pos);

    mat2x3 resultA  = mixMat2x3(fetchLighting(tex, pixelPos)              , fetchLighting(tex, pixelPos + ivec2(1, 0)), weights.x);
    mat2x3 resultB  = mixMat2x3(fetchLighting(tex, pixelPos + ivec2(0, 1)), fetchLighting(tex, pixelPos + ivec2(1, 1)), weights.x);

    return mixMat2x3(resultA, resultB, weights.y);
}


/* ------ ATROUS ------ */

ivec2 clampTexelPos(ivec2 pos) {
    return clamp(pos, ivec2(0.0), ivec2(viewSize));
}

float rgbHue(vec3 rgb) {
    float hue;
    if (rgb.x == rgb.y && rgb.y == rgb.z) hue = 0.0;
    else hue = (180.0 * rcp(pi)) * atan(2.0 * rgb.x - rgb.y - rgb.z, sqrt(3.0) * (rgb.y - rgb.z));

    if (hue < 0.0) hue = hue + 360.0;

    return clamp(hue, 0.0, 360.0);
}
float centerHue(float hue, float center) {
    float hueCentered = hue - center;

    if (hueCentered < -180.0) hueCentered += 360.0;
    else if (hueCentered > 180.0) hueCentered -= 360.0;

    return hueCentered;
}
float rgbSaturation(vec3 rgb) {
    float minrgb    = minOf(rgb);
    float maxrgb    = maxOf(rgb);

    return (max(maxrgb, 1e-10) - max(minrgb, 1e-10)) / max(maxrgb, 1e-2);
}

vec4 readHistoryGData(vec2 uv) {
    return texture(colortex10, uv);
}
vec4 readHistoryGData(ivec2 uv) {
    return texelFetch(colortex10, uv, 0);
}
float readLastDistance(ivec2 uv) {
    return texelFetch(colortex14, uv, 0).z;
}
vec4 readPreviousColor(ivec2 uv) {
    return texelFetch(colortex9, uv, 0);
}

float spatialEstimate(inout vec3 color, inout vec2 variance, vec2 uv, float centerDepth, vec3 centerNormal) {
    float sumWeight     = 1.0;
    float maxLum        = getLuma(color);

    ivec2 pixelUV       = ivec2(uv * viewSize);

    for (int x = -2; x <= 2; x++) {
        for (int y = -2; y <= 2; y++) {
            if (x == 0 && y == 0) continue;

            ivec2 pos   = pixelUV + ivec2(x, y);
            ivec2 posScale = ivec2(vec2(pos) / indirectResScale);

            float depth = depthLinear(texelFetch(depthtex0, posScale, 0).x);
            float depthDelta = distance(depth, centerDepth) * far;

            if (depthDelta < 2.0) {
                vec3 currentNormal  = decodeNormal(invF16UI(texelFetch(colortex1, posScale, 0).xy));

                float weight    = pow(max0(dot(centerNormal, currentNormal)), 16.0);

                vec3 currentColor   = texelFetch(colortex7, pos, 0).rgb;
                float currentLuma   = getLuma(currentColor);

                maxLum          = max(maxLum, currentLuma);
                color          += currentColor * weight;
                variance       += vec2(currentLuma * weight, sqr(currentLuma) * sqr(weight));
                sumWeight      += weight;
            }
        }
    }

    color       /= sumWeight;
    variance    /= vec2(sumWeight, sqr(sumWeight));

    return clamp(maxLum, 0.0, 2.71) * rpi;
}
vec3 spatialColor(vec2 uv, float centerDepth, vec3 centerNormal) {
    ivec2 pixelUV       = ivec2(uv * viewSize);
    vec3 color          = texelFetch(colortex7, pixelUV, 0).rgb;
    float sumWeight     = 1.0;

    for (int x = -2; x <= 2; x++) {
        for (int y = -2; y <= 2; y++) {
            if (x == 0 && y == 0) continue;

            ivec2 pos   = pixelUV + ivec2(x, y);
            ivec2 posScale = ivec2(vec2(pos) / indirectResScale);

            float depth = depthLinear(texelFetch(depthtex0, posScale, 0).x);
            float depthDelta = distance(depth, centerDepth) * far;

            if (depthDelta < 2.0) {
                vec3 currentNormal  = decodeNormal(invF16UI(texelFetch(colortex1, posScale, 0).xy));

                float weight    = pow(max0(dot(centerNormal, currentNormal)), 16.0) * exp(-depthDelta);

                vec3 currentColor   = texelFetch(colortex7, pos, 0).rgb;

                color          += currentColor * weight;
                sumWeight      += weight;
            }
        }
    }

    color       /= sumWeight;

    return color;
}

#include "/lib/offset/gauss.glsl"

vec3 spatialUpscale(sampler2D tex, sampler2D gData, vec3 sceneNormal, float centerDepth) {
    ivec2 pos           = ivec2(coord * viewSize / 4.0);

    vec4 centerData     = vec4(sceneNormal, sqr(centerDepth) * far);

    vec3 centerColor    = texelFetch(tex, pos, 0).rgb;

    vec3 total          = vec3(0);
    float totalWeight   = 0.0;

    for (int i = 0; i<25; i++) {
        ivec2 deltaPos      = kernelO_5x5[i];

        ivec2 samplePos     = pos + deltaPos;

        bool valid          = all(greaterThanEqual(samplePos, ivec2(0))) && all(lessThan(samplePos, ivec2(viewSize)));

        if (!valid) continue;

        vec4 currentData    = texelFetch(gData, clampTexelPos(samplePos), 0);
            currentData.xyz = currentData.xyz * 2.0 - 1.0;
            currentData.w   = sqr(currentData.w) * far;

        float depthDelta    = abs(currentData.w - centerData.w) / abs(centerData.w);

        float spatialWeight = pow(max(0.0, dot(currentData.xyz, centerData.xyz)), 8);
            spatialWeight  *= exp(-depthDelta * 4.0);

        vec3 currentColor   = texelFetch(tex, samplePos, 0).rgb;

        //accumulate stuff
        total       += currentColor * spatialWeight;

        totalWeight += spatialWeight;
    }

    //compensate for total sampling weight
    total          /= max(totalWeight, 1e-16);

    return total;
}

void main() {
    result          = vec4(0);
    accumulated     = vec4(0);
    GDataHistory    = vec4(0);
    varianceStore   = vec3(0);
    smoothColorStore = vec3(0);

    vec2 lowresCoord    = coord / indirectResScale;
    ivec2 pixelPos      = ivec2(floor(coord * viewSize) / indirectResScale);
    float sceneDepth    = texelFetch(depthtex0, pixelPos, 0).x;

    if (landMask(sceneDepth) && saturate(lowresCoord) == lowresCoord) {
        vec2 coord      = saturate(lowresCoord);
        vec2 scaledCoord = coord * indirectResScale;

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth));
        vec3 scenePos   = viewToSceneSpace(viewPos);

        bool hand       = sceneDepth < texelFetch(depthtex2, pixelPos, 0).x;

        vec3 reprojection   = reproject(scenePos, false);
        bool offscreen      = saturate(reprojection) != reprojection;

        vec2 scaledReprojection = reprojection.xy * indirectResScale;

        // Read Current Frame Geometry Data
        uvec4 tex1          = texelFetch(colortex1, ivec2(coord * viewSize), 0);
        vec3 currentNormal  = decodeNormal(invF16UI(tex1.xy));
        float currentDepth  = depthLinear(sceneDepth);
        float currentDistance = saturate(length(scenePos) / far);

        // Store Current Frame Geometry Data
        GDataHistory        = vec4(currentNormal * 0.5 + 0.5, sqrt(currentDepth));
        varianceStore.z     = sqrt(currentDistance);

        vec3 currentColor   = texture(colortex7, scaledCoord).rgb;
            currentColor    = spatialColor(scaledCoord, currentDepth, currentNormal);
        float currentLuma   = getLuma(currentColor);

        // Read Last Frame Geometry Data
        //vec3 lastNormal     = readHistoryGData(scaledReprojection).xyz * 2.0 - 1.0;
        //float lastDepth     = sqr(readHistoryGData(scaledReprojection).w);

        // Depth Rejection
        vec3 cameraMovement = mat3(gbufferModelView) * (cameraPosition - previousCameraPosition);
        //float depthDelta    = distance(lastDepth, currentDepth) - abs(cameraMovement.z / far);
        //bool depthRejection = depthDelta / abs(currentDepth) < 0.1;

        bool doSpatialEstimate = false;

        vec3 currentAdapt   = spatialUpscale(colortex15, colortex3, currentNormal, GDataHistory.w);

        if (offscreen) {
            doSpatialEstimate = true;
        } else {
            vec4 previousColor  = vec4(0);
            vec2 previousVariance = vec2(0);
            vec3 previousAdapt  = vec3(0);

            // Sample History
            ivec2 repPixel  = ivec2(floor(scaledReprojection * viewSize - vec2(0.5)));
            vec2 subpix     = fract(scaledReprojection * viewSize - vec2(0.5) - repPixel);

            const ivec2 offset[4] = ivec2[4](
                ivec2(0, 0),
                ivec2(1, 0),
                ivec2(0, 1),
                ivec2(1, 1)
            );

            float weight[4]     = float[4](
                (1.0 - subpix.x) * (1.0 - subpix.y),
                subpix.x         * (1.0 - subpix.y),
                (1.0 - subpix.x) * subpix.y,
                subpix.x         * subpix.y
            );

            float sumWeight     = 0.0;

            for (uint i = 0; i< 4; i++) {
                ivec2 pos   = repPixel + offset[i];

                float lastDepth     = sqr(readLastDistance(pos));

                float depthDelta    = distance(lastDepth, currentDistance) - abs(cameraMovement.z / far);
                bool depthRejection = depthDelta / abs(currentDistance) < 0.1;

                if (depthRejection) {
                    previousColor  += readPreviousColor(pos) * weight[i];
                    previousVariance += texelFetch(colortex14, pos, 0).xy * weight[i];
                    previousAdapt  += texelFetch(colortex11, pos, 0).rgb * weight[i];
                    sumWeight      += weight[i];
                }
            }

            if (sumWeight > 1e-3) {
                previousColor      /= sumWeight;
                previousVariance   /= sumWeight;
                previousAdapt      /= sumWeight;

                float frames        = min(previousColor.w * maxFrames + 1.0, maxFrames);
                float alphaColor    = max(0.025 * minAccumMult, 1.0 / frames);
                float alphaVariance = max(0.05 * minAccumMult, 1.0 / frames);

                #ifdef ACCUM_ADAPTIVE
                float currentHue    = rgbHue(currentAdapt);
                float historyHue    = rgbHue(previousAdapt);
                float hueDelta      = centerHue(historyHue, currentHue);

                float currentSaturation     = rgbSaturation(currentAdapt);
                float historySaturation     = rgbSaturation(previousAdapt);
                float saturationDelta       = (abs(currentSaturation - historySaturation));

                float currentLuma   = avgOf(currentAdapt);
                float historyLuma   = avgOf(previousAdapt);
                float lumaDelta     = sqr(distance(currentLuma, historyLuma) / (mix(currentLuma, historyLuma, 0.5) + 1e-3));
                    lumaDelta      *= sqr(1.0 - (historySaturation));

                float colorRejection = 1.0 / (1.0 + sqr((max0(abs(hueDelta) - 27.0) / 90.0) * historySaturation) + saturationDelta * historySaturation * euler + max0(lumaDelta));

                float adaptWeight   = saturate(colorRejection);
                    adaptWeight     = mix(1.0 - adaptWeight, 0.0, 1.0 - exp(-length(cameraMovement) * euler));

                    alphaColor      = max(saturate(0.66 * adaptWeight * ADAPT_STRENGTH), alphaColor);
                    alphaVariance   = max(saturate(0.75 * adaptWeight * ADAPT_STRENGTH), alphaVariance);
                #else
                const float adaptWeight     = 1.0;
                #endif

                vec2 currentVar     = vec2(currentLuma, sqr(currentLuma));
                vec2 variance       = mix(previousVariance, currentVar, alphaVariance);

                vec3 color          = mix(previousColor.rgb, currentColor, alphaColor);

                result              = vec4(color, sqrt(max0(variance.y - sqr(variance.x))));
                result.a           *= max(1.0, 4.0 / frames);

                accumulated         = vec4(result.rgb, frames / maxFrames);


                //result.rgb          = vec3(colorRejection);

                varianceStore.xy    = variance;

                smoothColorStore    = mix(previousAdapt, currentAdapt, max(0.2, 1.0 / frames));
            } else {
                doSpatialEstimate = true;
            }
        }

        if (doSpatialEstimate) {
            result.a        = spatialEstimate(result.rgb, varianceStore.xy, scaledCoord, currentDepth, currentNormal);
            accumulated     = vec4(result.rgb, 0.0);

            smoothColorStore    = currentAdapt;
        }
    }

    accumulated     = clamp16F(accumulated);
    result          = clamp16F(result);
}