#version 400 compatibility

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

layout(location = 0) out vec4 data0;
layout(location = 1) out vec4 data1;

in vec2 coord;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform mat4 shadowModelView, shadowProjection;
uniform mat4 shadowModelViewInverse, shadowProjectionInverse;

#define FUTIL_MAT16
#include "/lib/fUtil.glsl"
#include "/lib/light/warp.glsl"
#include "/lib/atmos/waterConst.glsl"



vec3 shadowScreenToView(vec3 position, float warp) {
    position    = position * 2.0 - 1.0;
    position.z /= 0.2;
    position.xy *= warp;
    position    = projMAD(inverse(shadowProjection), position);

    return position;
}
vec3 shadowViewToScene(vec3 position) {
    position     = transMAD(inverse(shadowModelView), position);

    return position;
}

void main() {
    data0   = stex(shadowcolor0);
    data1   = stex(shadowcolor1);
    data1.yzw = vec3(0.0);

    if (coord.y > 0.5) {
        vec4 albedo     = texture(shadowcolor0, coord);
        vec2 depth      = vec2(texture(shadowtex0, coord).x, texture(shadowtex1, coord).x);

        data0           = vec4(vec3(0.25), 0.0);
        if (depth.x < depth.y) data0 = albedo * vec4(vec3(0.25), 1.0);

        if (stex(shadowcolor1).x > 0.5) {     //Water
            vec2 uv     = coord.yx;
                uv.x    = uv.x * 2.0 - 1.0;
            float warp      = calculateWarp(uv * 2.0 - 1.0);

            vec3 viewPos0   = shadowScreenToView(vec3(uv, depth.x), warp);
            vec3 scenePos0  = shadowViewToScene(viewPos0);
            vec3 viewPos1   = shadowScreenToView(vec3(uv, depth.y), warp);
            vec3 scenePos1  = shadowViewToScene(viewPos1);

            float surfaceDist = distance(scenePos0, scenePos1);

            float caustic   = albedo.r * 4.0;

            vec3 absorb     = exp(-max0(surfaceDist * waterDensity) * waterAbsorbCoeff);
                absorb     *= mix(caustic, 1.0, exp(-max0(surfaceDist) / euler));

            data0           = vec4(absorb * 0.25, 1.0);
        }
        data1.x = 0.0;
    }
}